
#include <queue>
#include <iostream>
#include <cassert>
#include <memory>
#include <cstdlib>

using namespace std;

#include "trie.h"

/*****************************************************************************************************************/
TrieNode::TrieNode(NodeType t)
{
	Type = t;
	match = NULL;
	SetFail(NULL);
}

NodeType TrieNode::GetType()
{
	return Type;
}

void TrieNode::SetFail(TrieNode *f)
{
	fail = f;
}

TrieNode *TrieNode::GetFail()
{
	return fail;
}

void TrieNode::SetMatch(TrieMatch *m)
{
	match = m;
}

TrieMatch *TrieNode::GetMatch()
{
	return match;
}
/*****************************************************************************************************************/

TrieMatch::TrieMatch() : TrieNode(Match)
{
	SetMatch(this); // We are our own match.
}

TrieNode *TrieMatch::Advance(unsigned c)
{
	static_cast<void>(c);
	return NULL;
}

TrieIterator *TrieMatch::GetIterator() const
{
	return new TrieNullIterator(this);
}

/*****************************************************************************************************************/

TrieVector::TrieVector() : TrieNode(Vector)
{
	for (int i=0; i < 256; ++i)
		vector[i]=NULL;
}

TrieNode *TrieVector::Advance(unsigned c)
{
	return vector[c];
}

TrieIterator *TrieVector::GetIterator() const
{
	return new TrieVectorIterator(this);
}

void TrieVector::SetTransition(unsigned c, TrieNode *target)
{
	vector[c] = target;
}

void TrieVector::Iterate( Functor &f )
{
	int i;
	for (i=0; i < 256; ++i )
	{
		TransitionMap	m = { (unsigned char)i, vector[i] };
		f( m );
	}
}

TrieVectorIterator::TrieVectorIterator( const TrieVector *tv )
{
	m_Tv = tv;
	First();
}

void TrieVectorIterator::First()
{
	m_Index = 0;
}

void TrieVectorIterator::Next()
{
	if (!isDone())
		m_Index++;
};

bool TrieVectorIterator::isDone() const
{
	if (m_Index < 256)
		return false;
	else
		return true;
}

TransitionMap& TrieVectorIterator::Current() const
{
	if ( !isDone() )
	{
		m_TMap.c = (unsigned char)m_Index;
		m_TMap.dest = m_Tv->vector[m_Index];
		return m_TMap;
	}
	else
	{
		throw out_of_range("Iterator out of range - Current not valid");
	}
};



/*****************************************************************************************************************/


TrieDigit::TrieDigit(unsigned char d) : TrieNode(Digit)
{
	digit = d;
	dest=NULL;
};

TrieNode *TrieDigit::Advance(unsigned c)
{
	return c == digit ? dest : NULL;
}

TrieIterator *TrieDigit::GetIterator() const
{
	return new TrieDigitIterator(this);
}

void TrieDigit::SetTransition(unsigned char c, TrieNode *target)
{
	digit = c;
	dest = target;
}

void TrieDigit::Iterate( Functor &f )
{
	TransitionMap m = { digit, dest };
	f( m );
}

TrieDigitIterator::TrieDigitIterator( const TrieDigit *td )
{
	index = 0;
	m_Td = td;
}

void TrieDigitIterator::First()
{
	index = 0;
}

void TrieDigitIterator::Next()
{
	if (!isDone())
		index++;
};

bool TrieDigitIterator::isDone() const
{
	if ( index==0)
		return false;
	else
		return true;
}

TransitionMap& TrieDigitIterator::Current() const
{
	if ( !isDone() )
	{
		m_TMap.c = (unsigned char)m_Td->digit;
		m_TMap.dest = m_Td->dest;
		return m_TMap;
	}
	else
	{
		throw out_of_range("Iterator out of range - Current not valid");
	}

}


/*****************************************************************************************************************/

Trie::Trie() : Current(0), Duplicates(0)
{
	Root = new TrieVector;
}

Trie::~Trie()
{
	queue<TrieNode*>	que;


	// This must be done on Root first
	auto_ptr< TrieIterator >  k( Root->GetIterator() );
	k->First();
	while (!k->isDone())
	{
		TransitionMap	&m = k->Current();
		if ( m.dest && m.dest != Root )
		{
			que.push(m.dest);
		}
		k->Next();
	}

	while ( que.empty() == false  ) // Process items in the queue
	{
		TrieNode *n = que.front();
		que.pop();

		auto_ptr< TrieIterator > i( n->GetIterator() );
		i->First();
		while (!i->isDone())
		{
			TransitionMap	&map = i->Current();
			if (map.dest)
			{
				/* Does it have a hanging match or is a terminal TrieMatch? */
				TrieMatch *m = map.dest->GetMatch();
				if ( m != map.dest )
					que.push(map.dest);
				if ( m != NULL )
					delete m;
			}
			i->Next();
		}
		delete n;
	}
	delete Root;
}

void Trie::SetTransition(TrieNode ***last_link, unsigned char c)
{
	TrieVector *v;
	TrieDigit *d;
	TrieMatch *m;

	// If NULL
	if ( **last_link == 0 )
	{
		// Create a new digit node
		TrieDigit *d = new TrieDigit(c);
		**last_link = d; // Add it into the trie;
		*last_link = &d->dest; // Propose new transition path
	}
	else
	{
		switch ( (**last_link)->GetType() )
		{
		case Digit:
			{
				d = static_cast<TrieDigit *>(**last_link);
				if ( d->digit == c )
				{
					*last_link = &d->dest; // Propose new transition path
				}
				else // Convert digit to a vector
				{
					v = new TrieVector();

					// Copy from digit structure
					v->vector[ d->digit ] = d->dest;
					v->SetMatch( d->GetMatch() );
					delete d;

					**last_link = v; // Put v into the trie;
					// Propose new transition path
					*last_link = &v->vector[ c ] ;
				}
			}
			break;

		case Vector:
			{
				v = static_cast<TrieVector *>(**last_link);
				*last_link = &v->vector[c];
			}
			break;
		case Match:
			{
				m = static_cast<TrieMatch *>(**last_link);
				// Create a new digit, and move the match into the digit structure.

				TrieDigit *d = new TrieDigit(c); // Create a new digit node
				d->SetMatch( m ); // Copy the match over
				**last_link = d; // Replace d into the trie;
				*last_link = &d->dest; // Propose new transition path
			}
			break;
		default:
			abort();

		}
	}


}

void Trie::Add(const Pattern *pat)
{
	size_t i=0;
	TrieNode *n = Root;
	TrieNode **last_link = &n;
	unsigned char c;
	while( i < pat->m_Len )
	{
		c = pat->m_Pattern[ i ];
		SetTransition(&last_link, c);
		// Move along
		++i;
	}

	TrieMatch *m = new TrieMatch();
	m->patterns.push_back(pat);
	if ( *last_link == 0 ) // End of trie?
	{
		*last_link = m;
	}
	else if ( (*last_link)->GetType() != Match )  // No, then add it to the ancilliary match field; - if it's not a duplicate, that is.
	{
		(*last_link)->SetMatch( m );
	}
	else // Throw it away then since it must be a duplicate signature
	{
		delete m;
		Duplicates++;
		cerr << "Database: " << pat->m_Name << " duplicates "
			<< (*last_link)->GetMatch()->patterns.back()->m_Name << endl;
	}

}

TrieNode *Trie::FindFail( TrieNode *parent, unsigned short symbol )
{
	TrieNode	*f = parent->GetFail();
	TrieNode	*x = NULL;
	while ( !x ) // Round and round, until a non failure is found. Root is always a non failure.
	{
		x = f->Advance( symbol );
		f = f->GetFail();
	}
	return x;
}

void Trie::Initialise()
{

	clog << "Enter Trie::Initialise()" << endl;

	queue<TrieNode*>	node_queue;

	Root->SetFail(NULL); // Never will we fail on the root, since all paths are made checked. (See below)
	struct Rit : Functor
	{
		queue<TrieNode*>	*Q;
		TrieNode			*Root;

		Rit( queue<TrieNode*> *aQ, TrieNode *aRoot ) : Q(aQ), Root(aRoot) {}

		virtual void operator()( const TransitionMap &m )
		{
			if ( m.dest )
			{
				m.dest->SetFail(Root);
				Q->push( m.dest );
			}
			else
			{
				Root->SetTransition( m.c, Root ); // All failures lead back to root.
			}
		}
	} rit(&node_queue, Root);
	Root->Iterate( rit );

	clog << "Trie::Initialise() finished pushing base" << endl;

	while ( node_queue.empty() == false  ) // Process items in the queue
	{
		TrieNode *n = node_queue.front();
		node_queue.pop();

		struct Levit : Functor
		{
			queue<TrieNode*>	*q;
			TrieNode			*n;
			Trie				*t;

			Levit( queue<TrieNode*> *aQ, TrieNode *aN, Trie *aT ) : q(aQ), n(aN),t(aT) {}

			virtual void operator()( const TransitionMap &m )
			{
				if ( m.dest )
				{
					TrieNode *child = m.dest;
					child->SetFail( t->FindFail(n, m.c) );

					/* Fix to propergate TrieMatches to places where there is a substring */
					TrieMatch *p_match = n->GetFail()->GetMatch();	/* Get parent match */
					if ( p_match != NULL )
					{
						TrieMatch	*c_match;
						c_match = child->GetMatch();
						if ( c_match == 0 )
						{
							c_match = new TrieMatch;
							child->SetMatch( c_match );
						}
						list<const Pattern*>::iterator i;
						for (i=p_match->patterns.begin(); i != p_match->patterns.end(); ++i)
						{
							c_match->patterns.push_back(*i);
						}
					}
					q->push(child);
				}
			}
		} levit( &node_queue, n, this);
		n->Iterate( levit );

	}

	clog << "Exit Trie::Initialise()" << endl;

}

void Trie::Reset()
{
	Current = Root;
}

bool Trie::Find(BlockBuffer *b, list<const Pattern*> **patterns)
{
	bool ret = false;
	TrieMatch	*m;
	*patterns = NULL;

	for (int c = b->data[b->i]; b->i < b->n && ret == false;  b->i++, c = b->data[b->i] )
	{
		assert( Current != NULL );
		TrieNode *next = Current->Advance(c);
		while (!next)
		{
			Current = Current->GetFail();
			next = Current->Advance(c);
		}
		Current = next;

		if ( (m = Current->GetMatch()) != NULL )
		{
			*patterns=&m->patterns;
			ret = true;
		}
	}

	return ret;
}
