#if !defined (TRIE_H)
#define TRIE_H

#include <string>
#include <list>
#include <stdexcept>

using namespace std;

#include "buffer.h"

struct Pattern
{
	string			m_Name;
	unsigned char	*m_Pattern;
	size_t			m_Len;
};

class Trie;

enum NodeType { Match, Vector, Digit };

class TrieMatch;
class TrieNode;
class TrieDigit;
class TrieVector;

struct TransitionMap
{
	unsigned char	c;
	TrieNode		*dest;
};

struct Functor
{
	virtual void operator()( const TransitionMap &t ) = 0;
};

class TrieIterator
{
public:
	virtual void First() = 0;
	virtual void Next() = 0;
	virtual bool isDone() const = 0;
	virtual TransitionMap& Current() const = 0;
};

class TrieNullIterator : public TrieIterator
{
public:
	TrieNullIterator( const TrieNode *tv ) { (void)tv; }
	virtual void First() {}
	virtual void Next() {}
	virtual bool isDone() const { return true; }
	virtual TransitionMap& Current() const { throw runtime_error("Iterator out of range - Current not valid"); }
};

class TrieVectorIterator : public TrieIterator
{
public:
	TrieVectorIterator( const TrieVector *tv );
	virtual void First();
	virtual void Next();
	virtual bool isDone() const;
	virtual TransitionMap& Current() const;
private:
	size_t					m_Index;
	const TrieVector		*m_Tv;
	mutable	TransitionMap	m_TMap;
};


class TrieDigitIterator : public TrieIterator
{
public:
	TrieDigitIterator( const TrieDigit *td );
	virtual void First();
	virtual void Next();
	virtual bool isDone() const;
	virtual TransitionMap& Current() const;
private:
	int						index;
	const TrieDigit			*m_Td;
	mutable TransitionMap	m_TMap; // The composited object to return as a reference.
};

class TrieNode
{
public:
	NodeType GetType();
	virtual void SetFail(TrieNode *f);
	virtual TrieNode *GetFail();
	virtual void SetMatch(TrieMatch *m);
	virtual TrieMatch *GetMatch();
	virtual TrieIterator *GetIterator() const = 0;
	virtual void Iterate( Functor &f ) = 0;
	virtual TrieNode *Advance(unsigned c) = 0;
	virtual void SetTransition(unsigned c, TrieNode *target) { (void)c; (void)target; }
protected:
	TrieNode(NodeType t);
private:
	NodeType	Type;
	TrieNode	*fail;
	TrieMatch	*match;
};

class TrieMatch : public TrieNode
{
public:
	TrieMatch();
	virtual TrieNode *Advance(unsigned c);
	virtual TrieIterator *GetIterator() const;
	virtual void Iterate( Functor &f ) {(void)f;}	/* No iteration */
//private:
	list<const Pattern *>	patterns;
};

class TrieVector : public TrieNode
{
private:
protected:
public:
	TrieNode		*vector[256];

	TrieVector();

	virtual TrieNode *Advance(unsigned c);
	virtual TrieIterator *GetIterator() const;
	virtual void SetTransition(unsigned c, TrieNode *target);
	virtual void Iterate( Functor &f );
};


class TrieDigit : public TrieNode
{
public:
	unsigned char digit;
	TrieNode *dest;

	TrieDigit(unsigned char d);

	virtual TrieNode *Advance(unsigned c);
	virtual TrieIterator *GetIterator() const;
	virtual void SetTransition(unsigned char c, TrieNode *target) ;
	virtual void Iterate( Functor &f );
};



class Trie
{
private:
	TrieVector			*Root;
	TrieNode			*Current;
	unsigned long int	Duplicates;
	void SetTransition(TrieNode ***last_link, unsigned char c);


public:
	Trie();
	~Trie();
	void Add(const Pattern *sig);
	void Reset();
	bool Find(BlockBuffer *b, list<const Pattern*> **patterns);
	void Initialise();
	unsigned long GetDuplicates() const { return Duplicates; }
	TrieNode * FindFail( TrieNode *parent, unsigned short symbol );
};



#endif /* !defined (TRIE_H) */
